%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                   %
% Interpreter for the linear logic with subexponentials (SELLF)     %
%                                                                   %
% Author: Giselle Machado N. Reis                                   %
%                                                            2010   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% OBS: Since there are a lot of calls to functions that are not acctually part of the rule,
%% it was conventioned that these calls be placed one tab ahead from the actual implementation
%% of the rule.

%% OBS2: the integer N in each rule is for debugging purposes.

%% OBS3: the integer H in each rule is the limit of the height of the proof.
%% "when doing focused proof search, one only need to use as the height of a proof the number of decide rules encountered (instead of all introduction rules)" Dale Miller's e-mail

module sellf.

%%%% Inference rules for SELLF

sellf K G L :- prove_a 10000 K G L 0.

%%% Asynchronous Phase

prove_a H K G (top::L) N :- 
        printf [nl, pc " Top rule.", nl, fl],
        print_context K G.
            
prove_a H K G (bottom::L) N :- 
%        printf [nl, pc N, pc " Bottom rule: continuing with list '", pc L, nl, fl],
        N1 is N + 1,
    prove_a H K G L N1.
            
prove_a H K G ((A & B)::L) N :-
%        printf [nl, pc N, pc " Add and rule: divided in two branches with '", pc A, pc "' and '", pc B, pc "'", nl, fl],
        N1 is (N * 100 + 1),
        N2 is (N * 100 + 2),
    prove_a H K G (A::L) N1, prove_a H K G (B::L) N2.
            
prove_a H K G ((A @ B)::L) N :-
%        printf [nl, pc N, pc " Mult or rule: '", pc A, pc "' and '", pc B, pc "' put in the beginning of the list.", nl, fl], 
        N1 is N + 1, 
    prove_a H K G (A::B::L) N1.
            
prove_a H K G ((? I A)::L) N :-
    k_insert I A K K_new, 
%        printf [nl, pc N, pc " Question mark rule: inserting '", pc A, pc "' in subexp '", pc I, pc "'", nl, fl], 
        N1 is N + 1,
    prove_a H K_new G L N1.
            
prove_a H K G ((pi A)::L) N :-
%        printf [nl, pc N, pc " For all rule: with expression '", pc A, pc "'", nl, fl], 
        N1 is N + 1,
    pi X\ prove_a H K G ((A X)::L) N1.
        
%% Location creation
% TODO: test
prove_a H K G ((create I A)::L) N :-
%        printf [nl, pc N, pc  " Creating new subexponential index: ", pc I, nl, fl], 
        N1 is N + 1,
    prove_a H ((k_elm I nil)::K) G (A::L) N1.

%%% Synchronous Phase

prove_s H K nil (true) N :- 
        printf [nl, pc " True rule.", nl, fl],
    condition1 K. % side condition: K[I-W] is empty
            
prove_s H K G (A ; B) N :- 
%        printf [nl, pc N, pc " Add or 1 rule: chose to prove'", pc A, pc "'", nl, fl], 
        N1 is N + 1,
    prove_s H K G A N1.
            
prove_s H K G (A ; B) N :- 
%        printf [nl, pc N, pc " Add or 2 rule: chose to prove '", pc B, pc "'", nl, fl],
        N1 is N + 1, 
    prove_s H K G B N1.
            
prove_s H K G (A , B) N :-  
%        printf [nl, pc N, pc " Mult and rule: proving '", pc A, pc "' and '", pc B, pc "'", nl, fl],
    k_split K K1 K2,
    split G G1 G2,
    condition_and K1 K2,
        N1 is (N * 100 + 1),
        N2 is (N * 100 + 2),
    prove_s H K1 G1 A N1,
    prove_s H K2 G2 B N2. %% side condition: K1 and K2 must be equal in indexes that can suffer contraction and weakening
            
prove_s H K G (sigma A) N :-
%        printf [nl, pc N, pc " Exists rule: with expression '", pc A, pc "'", nl, fl],
        N1 is N + 1,
    sigma X\ prove_s H K G (A X) N1.
     
prove_s H K nil (bang I A) N :-
%% side condition: Every set with index not bigger than L (lower than or unrelated) which can't suffer weakening must be empty
    condition_bang I K, 
    k_geq I K K1,
%        printf [nl, pc N, pc " Bang rule: with term '", pc A, pc "' and subexp '", pc I, pc "'", nl, fl], 
        N1 is N + 1,
    prove_a H K1 nil (A::nil) N1. 
        
% making things easier =)
prove_s H K nil (bang_empty I A) N :-
    is_empty I K,
%        printf [nl, pc N, pc " Bang empty rule: with term '", pc A, pc "' and subexp '", pc I, pc "'", nl, fl], 
        N1 is N + 1,
    prove_a H K nil (A::nil) N1.

%% Location instantiation
%TODO: test 
prove_s H K G (inst I A) N :- 
%        printf [nl, pc N, pc " Instantiating subexponential index: ", pc I, nl, fl], 
        N1 is N + 1,
    sigma X\ (is_index X K, prove_s H K G ((I\ A) X) N1).

%%% Reaction, Identity and Decide rules

%% R - arrow down
prove_s H K G N Num:- 
    neg N, 
    %% XXX: commenting this term_to_string makes the program loop =(
        term_to_string N Str, 
%        printf [nl, pc " R arrow down rule: found '", pc N, pc "' negative, changing to asynchronous phase.", nl, fl], 
        Num1 is Num + 1,
    prove_a H K G (N::nil) Num1.

%% R - arrow up
prove_a H K G (P::L) N :- 
    (pos P ; literal P),
%        printf [nl, pc N, pc" R arrow up rule: passing positive or literal '", pc P, pc "' to linear (Gamma) context. Continuing with '", pc L, pc "'", nl, fl], 
        N1 is N + 1,
    prove_a H K (P::G) L N1.

%% D1
prove_a H K G nil N :- 
    insert_nd P G1 G,
    (pos P ; non_literal P),
    H > 0, H1 is H - 1,
%        printf [nl, pc N, pc " D1 rule: choosing '", pc P, pc "' positive from linear (Gamma) context, changing to synchronous phase.", nl, fl], 
        N1 is N + 1,
    prove_s H1 K G1 P N1.

%% Di - I can suffer contraction and weakening
prove_a H K G nil N :- 
    condition_dl I,
    k_insert I P K1 K,
    % Using this to avoid loops: everytime a formula is chosen, it is put in the end of the list.
    k_insert_end I P K1 K2,
    (pos P ; non_literal P), 
    H > 0, H1 is H - 1,
%        printf [nl, pc N, pc " Di 1 rule: choosing '", pc P, pc "' positive (or formula) from subexp '", pc I, pc "', changing to synchronous phase.", nl, fl],
        N1 is N + 1,
    prove_s H1 K2 G P N1. %% side condition: I can suffer contraction and weakening

%% Di - I can't suffer contraction or weakening
prove_a H K G nil N :- 
    not_condition_dl I, 
    k_insert I P K1 K,
    (pos P ; non_literal P),
    H > 0, H1 is H - 1,
%        printf [nl, pc N, pc " Di 2 rule: choosing '", pc P, pc "' positive (or formula) from subexp '", pc I, pc "', changing to synchronous phase.", nl, fl],
        N1 is N + 1,
    prove_s H1 K1 G P N1. %% side condition: I can't suffer contraction and weakening

%% I
prove_s H K G A N :- 
    condition_init K G A,
        printf [nl, pc " Initial rule: proving '", pc A, pc "'", nl, fl],
        print_context K G,
        flush std_out.

% Initial rule, version for when the atom is of the form l(m)
prove_s H K G (at I A) N :- 
    % (at I A) is positive and literal by definition
    condition_init_2 K G I A,
        printf [nl, pc " Initial rule v2: proving '", pc I, pc "(", pc A, pc ")", nl, fl],
        print_context K G,
        flush std_out.
        
% Initial rule, versions for arithmetical comparisson
prove_s H K nil (A <= B) N :-
    k_not_weakenable K nil, A <= B, 
        printf [nl, pc " Initial rule arith. comp. proving '", pc A, pc " <= ", pc B, pc "'", nl, fl],
        print_context K G,
        flush std_out.
prove_s H K nil (A < B) N :-
    k_not_weakenable K nil, A < B, 
        printf [nl, pc " Initial rule arith. comp. proving '", pc A, pc " < ", pc B, pc "'", nl, fl],
        print_context K G,
        flush std_out.
prove_s H K nil (A > B) N :-
    k_not_weakenable K nil, A > B, 
        printf [nl, pc " Initial rule arith. comp. proving '", pc A, pc " > ", pc B, pc "'", nl, fl],
        print_context K G,
        flush std_out.
prove_s H K nil (A >= B) N :-
    k_not_weakenable K nil, A >= B, 
        printf [nl, pc " Initial rule arith. comp. proving '", pc A, pc " >= ", pc B, pc "'", nl, fl],
        print_context K G,
        flush std_out.
% Initial rule, versions for arithmetical comparisson negated
prove_s H K nil (not (A <= B)) N :-
    k_not_weakenable K nil, not (A <= B), 
        printf [nl, pc " Initial rule arith. comp. proving 'not ", pc A, pc " <= ", pc B, pc "'", nl, fl],
        print_context K G,
        flush std_out.
prove_s H K nil (not (A < B)) N :-
    k_not_weakenable K nil, not (A < B), 
        printf [nl, pc " Initial rule arith. comp. proving 'not ", pc A, pc " < ", pc B, pc "'", nl, fl],
        print_context K G,
        flush std_out.
prove_s H K nil (not (A > B)) N :-
    k_not_weakenable K nil, not (A > B), 
        printf [nl, pc " Initial rule arith. comp. proving 'not ", pc A, pc " > ", pc B, pc "'", nl, fl],
        print_context K G,
        flush std_out.
prove_s H K nil (not (A >= B)) N :-
    k_not_weakenable K nil, not (A >= B), 
        printf [nl, pc " Initial rule arith. comp. proving 'not ", pc A, pc " >= ", pc B, pc "'", nl, fl],
        print_context K G,
        flush std_out.

%%%% Operations of the function K

% Erases the formulas in indexes that are lower than or not related to i.
k_geq I nil nil.
k_geq I ((k_elm X L)::T) ((k_elm X L)::T2) :- pred I X, k_geq I T T2.
k_geq I ((k_elm X L)::T) ((k_elm X nil)::T2) :- not (pred I X), k_geq I T T2.

% Chooses the formulas with subexponential index in S.
k_subset S nil nil.
k_subset S ((k_elm X L)::T) ((k_elm X L)::T2) :- is_in X S, k_subset S T T2.
k_subset S ((k_elm X L)::T) K_new :- not (is_in X S), k_subset S T K_new.   % in case X is not in S !is_in X S

% Inserts an element in one of K's list. Associates a new element with an existing subexponential index.
k_insert I A nil nil :- fail. % ERROR: this index does not exist.
k_insert I A ((k_elm I L)::T) ((k_elm I L1)::T) :- insert_nd A L L1.
k_insert I A ((k_elm X L)::T) ((k_elm X L)::T2) :- k_insert I A T T2.

% Union of functions
k_split nil nil nil.
k_split ((k_elm I L)::T) ((k_elm I L1)::T1) ((k_elm I L2)::T2) :- 
        not (cont I), 
        split L L1 L2,
        k_split T T1 T2.
% OBS: duplicating elements that can suffer weakening *and* contraction. Assuming that cont is subset of weak
k_split ((k_elm I L)::T) ((k_elm I L)::T1) ((k_elm I L)::T2) :- 
        cont I, 
        weak I, 
        k_split T T1 T2. 

% Gets the formulas with subexponential I (not in thesis - implemented to be used in condition_and)
k_get I nil nil :- fail.
k_get I ((k_elm I L)::T) L.
k_get I ((k_elm J L)::T) S :- k_get I T S.

% Returns the set K[I-W] (not in thesis - implemented to be used in condition_init)
k_not_weakenable nil nil.
k_not_weakenable ((k_elm I nil)::T) L1 :- not (weak I), k_not_weakenable T L1.
k_not_weakenable ((k_elm I L)::T) L :- not (weak I), k_not_weakenable T nil.
k_not_weakenable ((k_elm I L)::T) L2 :- weak I, k_not_weakenable T L2.

% Inserts an element in the end of one of K's list. (not in thesis - trying to solve loops)
k_insert_end I A nil nil :- fail. % ERROR: this index does not exist.
k_insert_end I A ((k_elm I L)::T) ((k_elm I L1)::T) :- append L (A::nil) L1.
k_insert_end I A ((k_elm X L)::T) ((k_elm X L)::T2) :- k_insert_end I A T T2.


%%%% Side Conditions' expressions

% K[I\W] = empty (only the formulas that can suffer weakening are left - exactly because they be weakened)
condition1 nil.
condition1 ((k_elm I L)::T) :- weak I, condition1 T.
condition1 ((k_elm I L)::T) :- L = nil, condition1 T.

% I is in C and W
condition_dl I :- cont I, weak I.

% I is not in C and W at the same time
not_condition_dl I :- weak I, not (cont I).
not_condition_dl I :- cont I, not (weak I).
not_condition_dl I :- not (cont I), not (weak I).

% K1 = K2 in subexponentials that can suffer contraction and weakening
condition_and nil _.
condition_and ((k_elm I L)::T) K :- 
        cont I, 
        weak I, 
        k_get I K L, 
        condition_and T K.
condition_and ((k_elm I L)::T) K :- 
% because I don't want this to be checked two times when both conditions are met.
        (not (cont I) ; not (weak I) ), !, 
        condition_and T K.

% Every set with index not bigger than L (lower than or unrelated) which can't suffer weakening must be empty
condition_bang L nil.
condition_bang L ((k_elm I nil)::T) :- not (pred L I), not (weak I), condition_bang L T. 
condition_bang L ((k_elm I F)::T) :- pred L I, condition_bang L T.
condition_bang L ((k_elm I F)::T) :- weak I, condition_bang L T.

% (not A) is in G union K[I] and G union K[I-W] is contained in {not A}  (for the initial axiom)
condition_init K G A :- 
            pos A, 
            literal A,
            ( G = ((my_not A)::nil) ; (is_in_k (my_not A) K, G = nil) ), 
            (k_not_weakenable K nil ; k_not_weakenable K ((my_not A)::nil)). 

condition_init_2 K G I A :- 
            has_index (my_not A) I K , G = nil , 
            (k_not_weakenable K nil ; k_not_weakenable K ((my_not A)::nil)). 
            
condition_init K G (my_not A) :- 
            pos (my_not A), 
            literal (my_not A),
            ( G = ((A)::nil) ; (is_in_k (A) K, G = nil) ), 
            (k_not_weakenable K nil ; k_not_weakenable K ((A)::nil)). 
            
condition_init_2 K G I (my_not A) :- 
            has_index A I K , G = nil , 
            (k_not_weakenable K nil ; k_not_weakenable K ((A)::nil)).     


%%%% Positive and negative formulas
pos (A , B).
pos (A ; B).
pos (bang I A).
pos (sigma A).
pos true.
pos zero.
pos (at I A). % l(m)
pos (my_not A) :- neg A.

pos (A <= B).
pos (A < B).
pos (A > B).
pos (A >= B).

neg (A & B).
neg (A @ B).
neg (? I A).
neg (pi A).
neg top.
neg bottom.
neg (my_not A) :- pos A.

%%%% Definition of literal
non_literal (A , B).
non_literal (A ; B).
non_literal (bang I A).
non_literal (sigma A).
non_literal (true).
non_literal (zero).
non_literal (A & B).
non_literal (A @ B).
non_literal (? I A).
non_literal (pi A).
non_literal (top).
non_literal (bottom).
literal A :- not (non_literal A).

%pred A A :- !. % reflexivity
%pred A C :- pred A B, not (A = B), pred B C, !. % transitivity (cuts are there to avoid infinite loop)

%%%% Auxiliar functions

% is_in X L -> true if the element X is in the list L.
is_in X (X::T).
is_in X (Y::T) :- is_in X T.

% is_in_k A K -> version of the funcion above for K lists
is_in_k A ((k_elm I L)::T) :- is_in A L.
is_in_k A ((k_elm I L)::T) :- is_in_k A T.

% true if X is a subexponential index in K
is_index X ((k_elm X L)::T).
is_index X ((k_elm I L)::T) :- is_index X T.

% true if A has index I in K
has_index A I ((k_elm I L)::T) :- is_in A L.
has_index A I ((k_elm X L)::T) :- has_index A I T.

% true if index I is empty is k
is_empty I ((k_elm I nil)::T).
is_empty I ((k_elm X L)::T) :- is_empty I T.

append nil K K.
append (H1::T1) K (H1::T2) :- append T1 K T2.

split nil nil nil.
split (H::L) (H::L1) L2 :- split L L1 L2.
split (H::L) L1 (H::L2) :- split L L1 L2.

% non-deterministic insertion
insert_nd A L (A::L).
insert_nd A (H::L) (H::L1) :- insert_nd A L L1.


%%%% I/O

printf nil.
printf (pc (Item : string) :: L) :- !, print Item, printf L.
printf (pc Item :: L)            :- term_to_string Item Str, print Str, printf L.
printf (nl :: L)                 :- print "\n", printf L.
printf (fl :: L)                 :- flush std_out, printf L.

print_context K G :- print_k K, print "\n", print_g G, flush std_out.

print_k K                    :- print "\nContext K: \n", print_k_rec K.
print_k_rec nil.
print_k_rec ((k_elm I L)::T) :- term_to_string I StrI, term_to_string L StrL,
                                print "- ", print StrI, print " = {", print StrL, print "}\n",
                                print_k_rec T.
                                
print_g G :- print "Context G = {", term_to_string G StrG, print StrG, print "}\n".


